/***********************************************************************

	This file is part of the Fuzzy Instance Based Learning package, a
	Java package implementing Fuzzy Nearest Neighbor Classifiers as 
	complementary material for the paper:
	
	Fuzzy Nearest Neighbor Algorithms: Taxonomy, Experimental analysis and Prospects

	Copyright (C) 2012
	
	J. Derrac (jderrac@decsai.ugr.es)
	S. Garca (sglopez@ujaen.es)
	F. Herrera (herrera@decsai.ugr.es)

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see http://www.gnu.org/licenses/
  
**********************************************************************/

/**
 * 
 * File: PosIBL.java
 * 
 * The PosIBL algorithm. 
 * 
 * @author Written by Joaqun Derrac (University of Granada) 13/11/2011 
 * @version 1.0 
 * @since JDK1.5
 * 
 */

package keel.Algorithms.Fuzzy_Instance_Based_Learning.PosIBL;

import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.Arrays;
import java.util.StringTokenizer;

import org.core.Files;

import keel.Algorithms.Fuzzy_Instance_Based_Learning.FuzzyIBLAlgorithm;
import keel.Algorithms.Fuzzy_Instance_Based_Learning.ReportTool;
import keel.Algorithms.Fuzzy_Instance_Based_Learning.Timer;
import keel.Algorithms.Fuzzy_Instance_Based_Learning.Util;

public class PosIBL extends FuzzyIBLAlgorithm {

	private double BETA;
	
	private double delta [];

	/** 
	 * Reads the parameters of the algorithm. 
	 * 
	 * @param script Configuration script
	 * 
	 */
	@Override
	protected void readParameters(String script) {
		
		String file;
		String line;
		String type;
		StringTokenizer fileLines, tokens;
		
	    file = Files.readFile (script);
	    fileLines = new StringTokenizer (file,"\n\r");
	    
	    //Discard in/out files definition
	    fileLines.nextToken();
	    fileLines.nextToken();
	    fileLines.nextToken();
	    
	    //Getting the M parameter
	    line = fileLines.nextToken();
	    tokens = new StringTokenizer (line, "=");
	    tokens.nextToken();
	    BETA = Double.parseDouble(tokens.nextToken().substring(1));
	    
	   
	    
	} //end-method	

	/**
	 * Main builder. Initializes the methods' structures
	 * 
	 * @param script Configuration script
	 */
	public PosIBL(String script){
		
		readDataFiles(script);
		
		//Naming the algorithm
		name="PosIBL";
		
		delta = new double [trainData.length];
		
			
		Arrays.fill(delta, 0.0);
		
	    //Initialization of Reporting tool
	    ReportTool.setOutputFile(outFile[2]);
	    
	} //end-method	
	
	/**
	 * Generates the model of the algorithm
	 */
	public void generateModel (){
		
		//Start of model time
		Timer.resetTime();	
		
		double minDist=0;
		double dist;
		
		//search for the nearest enemy
		for(int i=0;i<trainData.length;i++){
			minDist=Double.MAX_VALUE;
			for(int j=0;j<trainData.length;j++){
			
				if(trainOutput[i]!=trainOutput[j]){
					dist=Util.euclideanDistance(trainData[i], trainData[j]);
					
					if(minDist>dist){
						minDist=dist;
					}
					
				}
			
			}
			delta[i]=minDist;
			
			if(delta[i]<(1.0-BETA)){
				delta[i]=1.0-BETA;
			}
		}

		//End of model time
		Timer.setModelTime();
	
		//Showing results
		System.out.println(name+" "+ relation + " Model " + Timer.getModelTime() + "s");
		
	}
	
	/**
	 * Classifies the training set (leave-one-out)
	 */
	public void classifyTrain(){
	    
		//Start of training time
		Timer.resetTime();
		
		classifyTrainSet();
		
		//End of training time
		Timer.setTrainingTime();
		
		//Showing results
		System.out.println(name+" "+ relation + " Training " + Timer.getTrainingTime() + "s");
		
	} //end-method	
	
	/**
	 * Classifies the test set
	 */
	public void classifyTest(){
		    
		//Start of training time
		Timer.resetTime();
		
		classifyTestSet();
		
		//End of test time
		Timer.setTestTime();
		
		//Showing results
		System.out.println(name+" "+ relation + " Test " + Timer.getTestTime() + "s");	
		
	} //end-method	
	
	/**
	 * Classifies the training set
	 */
	public void classifyTrainSet(){
				
		for(int i=0;i<trainData.length;i++){
			
			trainPrediction[i]=computeClass(i);
			
		}

	} //end-method	
	
	/**
	 * Classifies the test set
	 */
	public void classifyTestSet(){

		for(int i=0;i<testData.length;i++){

			testPrediction[i]=computeTestClass(i);
			
		}

	} //end-method	
	
	/**
	 * Computes the class of a instance given its membership array
	 * @param pertenence Membership array
	 * 
	 * @return Class assigned (crisp)
	 */
	private int computeClass(int index){
		
		double maxSim=Double.MIN_VALUE;
		double dist,sim;
		int nearest=0;
		int result;
		
		for(int i=0;i<trainData.length;i++){
			
			if(i!=index){
				dist=Util.euclideanDistance(trainData[i], trainData[index]);
				
				//kernel function
				dist=1.0-dist;
				dist=-1.0*delta[i]*(1.0-dist);
				sim=Math.exp(dist);

				if(maxSim<sim){
					maxSim=sim;
					nearest=i;
				}
			}
		}
		
		result=trainOutput[nearest];
		
		return result;
		
	} //end-method	
	
	/**
	 * Computes the class of a instance given its membership array
	 * @param pertenence Membership array
	 * 
	 * @return Class assigned (crisp)
	 */
	private int computeTestClass(int index){
		
		double maxSim=Double.MIN_VALUE;
		double dist,sim;
		int nearest=0;
		int result;
		
		for(int i=0;i<trainData.length;i++){
			
			dist=Util.euclideanDistance(trainData[i], testData[index]);
				
			//kernel function
			dist=1.0-dist;
			dist= -1.0*delta[i]*(1.0-dist);
			sim=Math.exp(dist);
				
			if(maxSim<sim){
				maxSim=sim;
				nearest=i;
			}
		}
		
		result=trainOutput[nearest];
		
		return result;
		
	} //end-method	
	
	/**
	 * Reports the results obtained
	 */
	public void printReport(){
		
		writeOutput(outFile[0], trainOutput, trainPrediction);
		writeOutput(outFile[1], testOutput, testPrediction);
		
		ReportTool.setResults(trainOutput,trainPrediction,testOutput,testPrediction,nClasses);
		
		ReportTool.printReport();
		
	} //end-method	
    
} //end-class 
